@@ -1,7 +1,7 @@
-#ifndef __vpi_user_H
-#define __vpi_user_H
+#ifndef VPI_USER_H
+#define VPI_USER_H
 /*
- * Copyright (c) 1999-2010 Stephen Williams (steve@icarus.com)
+ * Copyright (c) 1999-2020 Stephen Williams (steve@icarus.com)
  *
  *    This source code is free software; you can redistribute it
  *    and/or modify it in source code form under the terms of the GNU
@@ -16,11 +16,11 @@
  *
  *    You should have received a copy of the GNU General Public License
  *    along with this program; if not, write to the Free Software
- *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 
-#if defined(__MINGW32__) || defined (__CYGWIN32__)
+#if defined(__MINGW32__) || defined (__CYGWIN__)
 #  define DLLEXPORT __declspec(dllexport)
 #else
 #  define DLLEXPORT
@@ -43,9 +43,17 @@ EXTERN_C_START
 
 # include  <stdarg.h>
 # include  <stdio.h>
+# include  <stdarg.h>
 # include  "_pli_types.h"
 
+#ifndef ICARUS_VPI_CONST
+#define ICARUS_VPI_CONST
+#endif
+#ifdef __cplusplus
+typedef class __vpiHandle *vpiHandle;
+#else
 typedef struct __vpiHandle *vpiHandle;
+#endif
 
 /*
  * This structure is created by the VPI application to provide hooks
@@ -55,10 +63,10 @@ typedef struct t_vpi_systf_data {
       PLI_INT32 type;
       PLI_INT32 sysfunctype;
       const char *tfname;
-      PLI_INT32 (*calltf)   (PLI_BYTE8*);
-      PLI_INT32 (*compiletf)(PLI_BYTE8*);
-      PLI_INT32 (*sizetf)   (PLI_BYTE8*);
-      PLI_BYTE8 *user_data;
+      PLI_INT32 (*calltf)   (ICARUS_VPI_CONST PLI_BYTE8*);
+      PLI_INT32 (*compiletf)(ICARUS_VPI_CONST PLI_BYTE8*);
+      PLI_INT32 (*sizetf)   (ICARUS_VPI_CONST PLI_BYTE8*);
+      ICARUS_VPI_CONST PLI_BYTE8 *user_data;
 } s_vpi_systf_data, *p_vpi_systf_data;
 
 /* The type in the above structure can have one of the following
@@ -223,7 +231,7 @@ typedef struct t_vpi_delay  {
       PLI_INT32  time_type; /* vpiScaledRealTime, vpiSimTime */
       PLI_INT32  mtm_flag;
       PLI_INT32  append_flag;
-      PLI_INT32  plusere_flag;
+      PLI_INT32  pulsere_flag;
 } s_vpi_delay, *p_vpi_delay;
 
 
@@ -278,15 +286,20 @@ typedef struct t_vpi_delay  {
 #define vpiNamedEvent  34
 #define vpiNamedFork   35
 #define vpiNet         36
+#define vpiNetBit      37
 #define vpiParameter   41
 #define vpiPartSelect  42
 #define vpiPathTerm    43
+#define vpiPort        44
 #define vpiRealVar     47
 #define vpiReg         48
+#define vpiRegBit      49
 #define vpiSysFuncCall 56
 #define vpiSysTaskCall 57
 #define vpiTask        59
 #define vpiTimeVar     63
+#define vpiUdpDefn     66
+#define vpiUserSystf   67
 #define vpiNetArray   114
 #define vpiIndex       78
 #define vpiLeftRange   79
@@ -301,7 +314,14 @@ typedef struct t_vpi_delay  {
 #define vpiVariables   100
 #define vpiExpr        102
 
-#define vpiCallback  1000
+/********************** object types added with 1364-2001 *********************/
+
+#define vpiCallback   107
+#define vpiRegArray   116
+
+/********************** object types added with 1364-2005 *********************/
+
+#define vpiGenScope   134
 
 /* PROPERTIES */
 #define vpiUndefined   (-1)
@@ -318,6 +338,16 @@ typedef struct t_vpi_delay  {
 #define vpiTimePrecision 12
 #define vpiDefFile       15
 #define vpiDefLineNo     16
+#define vpiScalar        17
+#define vpiVector        18
+
+#define vpiDirection 20 /* direction of port: */
+#   define vpiInput 1
+#   define vpiOutput 2
+#   define vpiInout 3
+#   define vpiMixedIO 4 /* Not currently output */
+#   define vpiNoDirection 5
+
 #define vpiNetType       22
 #   define vpiWire         1
 #   define vpiWand         2
@@ -331,6 +361,7 @@ typedef struct t_vpi_delay  {
 #   define vpiSupply1     10
 #   define vpiSupply0     11
 #define vpiArray         28
+#define vpiPortIndex     29
 #define vpiEdge          36
 #   define vpiNoEdge       0x00 /* No edge */
 #   define vpiEdge01       0x01 /* 0 --> 1 */
@@ -360,11 +391,20 @@ typedef struct t_vpi_delay  {
 #   define vpiSysFuncReal  vpiRealFunc
 #   define vpiSysFuncTime  vpiTimeFunc
 #   define vpiSysFuncSized vpiSizedFunc
+#define vpiUserDefn       45
 #define vpiAutomatic      50
 #define vpiConstantSelect 53
 #define vpiSigned         65
+#define vpiLocalParam     70
 /* IVL private properties, also see vvp/vpi_priv.h for other properties */
-#define _vpiNexusId 0x1000000
+#define _vpiNexusId        0x1000000
+/* used in vvp/vpi_priv.h  0x1000001 */
+#define _vpiDelaySelection 0x1000002
+#  define _vpiDelaySelMinimum 1
+#  define _vpiDelaySelTypical 2
+#  define _vpiDelaySelMaximum 3
+/* used in vvp/vpi_priv.h  0x1000003 */
+/* used in vvp/vpi_priv.h  0x1000004 */
 
 /* DELAY MODES */
 #define vpiNoDelay            1
@@ -377,17 +417,25 @@ typedef struct t_vpi_delay  {
 #define vpiReturnEvent 0x1000
 
 /* VPI FUNCTIONS */
-extern void vpi_register_systf(const struct t_vpi_systf_data*ss);
+extern vpiHandle vpi_register_systf(const struct t_vpi_systf_data*ss);
+extern void vpi_get_systf_info(vpiHandle obj, p_vpi_systf_data data);
 
 /* I/O routines */
 extern PLI_UINT32 vpi_mcd_open(char *name);
 extern PLI_UINT32 vpi_mcd_close(PLI_UINT32 mcd);
 extern char      *vpi_mcd_name(PLI_UINT32 mcd);
 extern PLI_INT32  vpi_mcd_printf(PLI_UINT32 mcd, const char*fmt, ...)
+#ifdef __MINGW32__
+      __attribute__((format (gnu_printf,2,3)));
+#else
       __attribute__((format (printf,2,3)));
-
+#endif
 extern PLI_INT32  vpi_printf(const char*fmt, ...)
+#ifdef __MINGW32__
+      __attribute__((format (gnu_printf,1,2)));
+#else
       __attribute__((format (printf,1,2)));
+#endif
 
 extern PLI_INT32  vpi_vprintf(const char*fmt, va_list ap);
 extern PLI_INT32  vpi_mcd_vprintf(PLI_UINT32 mcd, const char*fmt, va_list ap);
@@ -422,7 +470,7 @@ typedef struct t_cb_data {
       p_vpi_time time;
       p_vpi_value value;
       PLI_INT32 index;
-      char      *user_data;
+      ICARUS_VPI_CONST PLI_BYTE8 *user_data;
 } s_cb_data, *p_cb_data;
 
 #define cbValueChange        1
@@ -449,6 +497,7 @@ typedef struct t_cb_data {
 #define cbExitInteractive   22
 #define cbInteractiveScopeChange 23
 #define cbUnresolvedSystf   24
+#define cbAtEndOfSimTime    31
 
 extern vpiHandle vpi_register_cb(p_cb_data data);
 extern PLI_INT32 vpi_remove_cb(vpiHandle ref);
@@ -525,7 +574,10 @@ extern void vpi_get_delays(vpiHandle expr, p_vpi_delay delays);
 extern void vpi_put_delays(vpiHandle expr, p_vpi_delay delays);
 
 
-
+/*
+ * Check to see if two handles point to the same object.
+ */
+extern PLI_INT32 vpi_compare_objects(vpiHandle obj1, vpiHandle obj2);
 
 
 /*
@@ -581,9 +633,108 @@ extern DLLEXPORT void (*vlog_startup_routines[])(void);
   /* Format a scalar a la %v. The str points to a 4byte character
      buffer. The value must be a vpiStrengthVal. */
 extern void vpip_format_strength(char*str, s_vpi_value*value, unsigned bit);
+  /* Set the return value to return from the vvp run time. This is
+     usually 0 or 1. This is the exit code that the vvp process
+     returns, and in distinct from the finish_number that is an
+     argument to $fatal and other severity tasks. The $fatal and
+     $finish system tasks bundled with iverilog use this function to
+     tell vvp to exit SUCCESS or FAILURE. */
 extern void vpip_set_return_value(int value);
+
 extern s_vpi_vecval vpip_calc_clog2(vpiHandle arg);
+extern void vpip_make_systf_system_defined(vpiHandle ref);
+
+  /* Perform fwrite to mcd files. This is used to write raw data,
+     which may include nulls. */
+extern void vpip_mcd_rawwrite(PLI_UINT32 mcd, const char*buf, size_t count);
+
+  /* Return driver information for a net bit. The information is returned
+     in the 'counts' array as follows:
+       counts[0] - number of drivers driving '0' onto the net
+       counts[1] - number of drivers driving '1' onto the net
+       counts[2] - number of drivers driving 'X' onto the net
+       counts[3] - set to 1 if the net is forced, 0 otherwise
+     The 'ref' argument should reference a net. The 'idx' argument selects
+     which bit of the net is examined. */
+extern void vpip_count_drivers(vpiHandle ref, unsigned idx,
+                               unsigned counts[4]);
+
+/*
+ * Stopgap fix for br916. We need to reject any attempt to pass a thread
+ * variable to $strobe or $monitor. To do this, we use some private VPI
+ * properties that are normally only used by the VVP thread cleanup code.
+ * Normally the following definitions are provided by vvp/vpi_priv.h, but
+ * for the stopgap fix we need to make them more widely available.
+ */
+#define BR916_STOPGAP_FIX
+#ifdef BR916_STOPGAP_FIX
+#define _vpiFromThr 0x1000001
+#   define _vpiNoThr   0
+#   define _vpiString  1
+#   define _vpiVThr    2
+#   define _vpiWord    3
+#   define _vpi_at_PV  4
+#   define _vpi_at_A   5
+#   define _vpi_at_APV 6
+#endif
+
+#if defined(__MINGW32__) || defined (__CYGWIN__)
+/*
+ * In Linux, when loaded, a shared library can automatically bind to functions
+ * provided by its client. In Windows, a DLL can only do this statically at
+ * link time, and is then tied to a specific client. So to enable VPI modules
+ * to be used by both the compiler and the simulator, we construct a jump table
+ * for the VPI routines that we can pass down to the VPI modules.
+ */
+
+// Increment the version number any time vpip_routines_s is changed.
+static const PLI_UINT32 vpip_routines_version = 1;
+
+typedef struct {
+    vpiHandle   (*register_cb)(p_cb_data);
+    PLI_INT32   (*remove_cb)(vpiHandle);
+    vpiHandle   (*register_systf)(const struct t_vpi_systf_data*ss);
+    void        (*get_systf_info)(vpiHandle, p_vpi_systf_data);
+    vpiHandle   (*handle_by_name)(const char*, vpiHandle);
+    vpiHandle   (*handle_by_index)(vpiHandle, PLI_INT32);
+    vpiHandle   (*handle)(PLI_INT32, vpiHandle);
+    vpiHandle   (*iterate)(PLI_INT32, vpiHandle);
+    vpiHandle   (*scan)(vpiHandle);
+    PLI_INT32   (*get)(int, vpiHandle);
+    char*       (*get_str)(PLI_INT32, vpiHandle);
+    void        (*get_delays)(vpiHandle, p_vpi_delay);
+    void        (*put_delays)(vpiHandle, p_vpi_delay);
+    void        (*get_value)(vpiHandle, p_vpi_value);
+    vpiHandle   (*put_value)(vpiHandle, p_vpi_value, p_vpi_time, PLI_INT32);
+    void        (*get_time)(vpiHandle, s_vpi_time*);
+    void*       (*get_userdata)(vpiHandle);
+    PLI_INT32   (*put_userdata)(vpiHandle, void*);
+    PLI_UINT32  (*mcd_open)(char *);
+    PLI_UINT32  (*mcd_close)(PLI_UINT32);
+    PLI_INT32   (*mcd_flush)(PLI_UINT32);
+    char*       (*mcd_name)(PLI_UINT32);
+    PLI_INT32   (*mcd_vprintf)(PLI_UINT32, const char*, va_list);
+    PLI_INT32   (*flush)(void);
+    PLI_INT32   (*vprintf)(const char*, va_list);
+    PLI_INT32   (*chk_error)(p_vpi_error_info);
+    PLI_INT32   (*compare_objects)(vpiHandle, vpiHandle);
+    PLI_INT32   (*free_object)(vpiHandle);
+    PLI_INT32   (*get_vlog_info)(p_vpi_vlog_info info) ;
+    void        (*vcontrol)(PLI_INT32, va_list);
+    PLI_INT32   (*fopen)(const char*, const char*);
+    FILE*       (*get_file)(PLI_INT32);
+    s_vpi_vecval(*calc_clog2)(vpiHandle);
+    void        (*count_drivers)(vpiHandle, unsigned, unsigned [4]);
+    void        (*format_strength)(char*, s_vpi_value*, unsigned);
+    void        (*make_systf_system_defined)(vpiHandle);
+    void        (*mcd_rawwrite)(PLI_UINT32, const char*, size_t);
+    void        (*set_return_value)(int);
+} vpip_routines_s;
+
+extern DLLEXPORT PLI_UINT32 vpip_set_callback(vpip_routines_s*routines, PLI_UINT32 version);
+
+#endif // defined(__MINGW32__) || defined (__CYGWIN__)
 
 EXTERN_C_END
 
-#endif
+#endif /* VPI_USER_H */
